use quote::quote;
use syn::{parse_macro_input, parse_quote, DeriveInput, GenericParam, Generics};

#[proc_macro_derive(SharedFunctorAny)]
pub fn derive_shared_functor(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
    let input = parse_macro_input!(input as DeriveInput);
    let name = input.ident;
    let generics = add_clone_bounds(input.generics);
    let (impl_generics, ty_generics, where_clause) = generics.split_for_impl();
    let expanded = quote! {
        impl #impl_generics SharedFunctorAny for #name #ty_generics #where_clause {
            type SharedAny<'a, A: 'a + Clone> = Self::FAny<'a, A>
            where
                Self: 'a;

            fn share<'a, A: 'a + Clone>(fa: Self::FAny<'a, A>) -> Self::SharedAny<'a, A>
            where
                Self: 'a,
            {
                fa
            }

            fn unshare<'a, A: 'a + Clone>(sa: Self::SharedAny<'a, A>) -> Self::FAny<'a, A>
            where
                Self: 'a,
            {
                sa
            }
        }
    };
    proc_macro::TokenStream::from(expanded)
}

fn add_clone_bounds(mut generics: Generics) -> Generics {
    for param in &mut generics.params {
        if let GenericParam::Type(ref mut type_param) = *param {
            type_param.bounds.push(parse_quote!(Clone));
        }
    }
    generics
}
