//! Traversible binary trees.

pub mod algorithms;

use std::rc::Rc;

use crate::flow::comparator::*;
use crate::func::*;

pub type Split<'a, T, A, D> = (
    Rc<dyn TraversibleBinaryTree<'a, T, A, D>>,
    Rc<dyn TraversibleBinaryTree<'a, T, A, D>>,
    A,
);

pub trait TraversibleBinaryNode<'a, T: Monad<'a>, A: 'a, D: 'a + PartialEq>: 'a {
    fn split(&self) -> Split<'a, T, A, D>;

    fn to_tree(self: Rc<Self>) -> Rc<dyn TraversibleBinaryTree<'a, T, A, D>>;
}

pub trait TraversibleBinaryReference<'a, T: Monad<'a>, A: 'a, D: 'a + PartialEq>: 'a {
    fn resolve(&self) -> Wrap<'a, Rc<dyn TraversibleBinaryNode<'a, T, A, D>>, T>;

    /// This should be enough to compare reference for equality.
    fn data(&self) -> D;
}

pub trait TraversibleBinaryTree<'a, T: Monad<'a>, A: 'a, D: 'a + PartialEq>: 'a {
    fn refer(&self) -> Option<Rc<dyn TraversibleBinaryReference<'a, T, A, D>>>;
}
