use std::cmp::max;

use crate::func::{context::*, *};
use crate::rcore::*;

use super::*;

pub struct TestContextCounted;

impl<'a> FunctorContext<'a> for TestContextCounted {
    type T = CountedInstance;
}

impl<'a> FallibleCtx<'a> for TestContextCounted {
    type Fallible = instances::result::ResultFailOver<Self::T>;
}

impl<'a> Context<'a> for TestContextCounted {
    type _Tm = Self::T;

    type D = NoDiagnostic;

    type LookupError = TestLookupError<'a>;

    fn hash(s: &[u8]) -> Hash {
        TestContextPlain::hash(s)
    }
}

pub type CountedInstance = instances::effect::EffectInstance<usize>;

impl instances::effect::Effect for usize {
    fn e_pure() -> Self {
        0
    }

    fn e_seq(el: Self, er: Self) -> Self {
        max(el, er)
    }

    fn e_after(self, effect: Self) -> Self {
        self + effect
    }
}

pub type Counted<A> = instances::effect::WithEffect<A, usize>;

impl<A> Counted<A> {
    fn add(self, n: usize) -> Self {
        Counted {
            value: self.value,
            effect: self.effect + n,
        }
    }

    pub fn count(&self) -> usize {
        self.effect
    }
}

struct CountedInject;

impl<'a> Inject<'a, TestContextCounted> for CountedInject {
    fn inject<A: 'a>(
        &self,
        fa: Wrapped<'a, TestContextCounted, A>,
    ) -> Wrapped<'a, TestContextCounted, A> {
        fa.add(1)
    }
}

pub trait Delayable<'a>: Mentionable<'a, TestContextCounted> + Sized {
    fn delay(&self) -> ParseResultA<'a, TestContextCounted, Self> {
        Rc::new(CountedInject).inject_mentionable(self)
    }
}

impl<'a, A: Mentionable<'a, TestContextCounted>> Delayable<'a> for A {}
